function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp, ks] = driver_PhaseRetrieval(b, masks, method, params, phase, metric, delta, isplot, kappa)
% driver for Phase Retrieval problem.
% 
% INPUT:
% b: observations
% masks: masks
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X : all iterates
% F : function values of all iterates
% G : norm of gradient of all iterates
% T : accumulated computational time for iterations
% timecost : total computational time
% nf : number of function evaluations
% ng : number of gradient evaluations
% nR : number of retraction actions
% nH : number of Hessian actions
% nV : number of vector transport actions without parameters
% nVp : number of vector transport action with parameters
% 
% By Wen Huang

    %% default function handles
    
    if(metric == 1)
        fns.f = @(x)f_q(x, masks, b, kappa);
        fns.gf = @(x)gf_q(x, masks, b, kappa);
        fns.R = @R_Exp_q;
        fns.inpro = @inpro_q;
        fns.Hv = @Hv_q;
        fns.phi = @phi_q;
        fns.beta = @beta_q;
        fns.Tranv = @Tranv_q;
        if(method == 11)
            fns.Tranv = @TranvProj_q;
        end
        fns.invTranv = @invTranv_q;
        fns.TranH = @TranH_q;
        fns.rank1operator = @rank1operator_q;
        fns.operadd = @operadd_q;
    elseif(metric == 2)
        fns.f = @(x)f(x, masks, b);
        fns.gf = @(x)gf(x, masks, b);
        fns.hessianv = @(x, v)fhessianv(x, v, masks, b);
        fns.fhess = @(x, v)fhessianv(x, v, masks, b);
        fns.proj = @proj;
        fns.R = @R_Exp;
        fns.inpro = @inpro;
    end
    fns.STV = @STV;
    fns.TVaddTV = @TVaddTV;
    fns.IsStopped = @(X, F, G, T, params, fns)IsStopped(X, F, G, T, params, fns, delta);
    
    global isfixedrank

    X = {}; F = []; G = []; T = []; timecost = 0; nf = 0; ng = 0; nR = 0; nH = 0; nV = 0; nVp = 0;ks = [];
    
    x0 = params.x0;
    r = size(x0.U, 2);
    ks = ones(r, 1);
    while(r >= 1)
        fprintf('rank is %d\n', r);
%         [Xi, Fi, Gi, Ti, timecosti, nfi, ngi, nRi, nHi, nVi, nVpi] = callalgo(method, fns, params);
%         params.debug = 0;%%---
%         [Xi, Fi, Gi, Ti, timecosti, nfi, ngi, nRi, nHi, nVi, nVpi] = callalgo(3, fns, params);%%---
%         params.x0 = Xi{end};%%---
        [Xi, Fi, Gi, Ti, timecosti, nfi, ngi, nRi, nHi, nVi, nVpi] = callalgo(method, fns, params);
        Xend = Xi{end};
        clearvars -except ks Xend F Fi G Gi T Ti timecost timecosti ...
            nf nfi ng ngi nR nRi nH nHi nV nVi nVp nVpi r isfixedrank method params metric delta fns isplot;
        Xi{1} = Xend;
        X = {};
        % combine output information
        X = [X, Xi]; F = [F, Fi]; G = [G, Gi]; timecost = timecost + timecosti;
        if(length(T) > 0)
            T = [T, Ti + T(end)];
        else
            T = Ti;
        end
        nf = nf + nfi; ng = ng + ngi; nR = nR + nRi; nH = nH + nHi; nV = nV + nVi; nVp = nVp + nVpi;
        if(r > 1)
            ks(1 : r - 1, 1) = length(F) + 1;
        end
        if(isfixedrank == 1 || method == 13)
            break;
        end
        xf = Xi{end};
        r = size(xf.U, 2);
        if((Gi(end) < params.error) && (r == 1)) % obtain solution for phaselift
            break;
        end

        if(r == 1) %if at a local minimizer of rank 1 matrix, then we increase rank.
            x0 = xf;
            if(metric == 1 || metric == 2)
                [U, S, V] = svd(xf.U, 0);
                Sigma = S * S;
            else
                U = xf.U;
                Sigma = xf.D;
            end
            u = PhaseRetrieval_RQ(U, Sigma, b, masks);
            x0.U = [U, u];
            x0.D = diag([diag(Sigma), 0]);
            tilde_x = OneIterationSD(x0, masks, b);
            if(metric == 1 || metric == 2)
                x0.U = tilde_x.U * sqrt(tilde_x.D);
            else
                x0 = tilde_x;
            end
         else %Since the true solution is rank one for phaselift, we reduce rank if rank is bigger than 1
%             if(metric == 1 || metric == 2)
%                 [U, S, V] = svd(xf.U, 0);
%                 x0.U = U(:, 1 : end - 1) * S(1 : end - 1, 1 : end - 1);% / norm(S(1 : end - 1, 1 : end - 1), 'fro') * norm(S, 'fro');
%             else
%                 x0.U = xf.U(:, 1 : end - 1);
%                 x0.D = xf.D(1 : end - 1, 1 : end - 1);% / norm(xf.D(1 : end - 1, 1 : end - 1), 'fro') * norm(xf.D, 'fro');
%             end
            if(metric == 1 || metric == 2)
                [U, S, V] = svd(xf.U, 0);
%                 x0.U = U(:, 1 : end - 1) * S(1 : end - 1, 1 : end - 1) / norm(S(1 : end - 1, 1 : end - 1), 'fro') * norm(S, 'fro');
                normSdsqrk = norm(diag(S)) / sqrt(size(S, 1));
                for i = 1 : size(U, 2)
                    % S(i, i) / S(1, 1)
                    if(S(i, i) / normSdsqrk < delta)
                        x0.U = U(:, 1 : i - 1) * S(1 : i - 1, 1 : i - 1);% / norm(S(1 : end - 1, 1 : end - 1), 'fro') * norm(S, 'fro');
                        break;
                    end
                end
            else
                for i = 1 : size(U, 2)
                    if(S(i, i) / S(1, 1) < delta)
                        x0.U = xf.U(:, 1 : i - 1);
                        x0.D = xf.D(1 : i - 1, 1 : i - 1);% / norm(xf.D(1 : end - 1, 1 : end - 1), 'fro') * norm(xf.D, 'fro');
                        break;
                    end
                end
            end
        end
        % get next initial point
        params.x0 = x0;
        r = size(x0.U, 2);
    end
    if(method == 13)
        return;
    end
    fprintf('Num. of total Iter.: %d, time cost: %f seconds, Num. of Fun, Gra, Hv, VT and R: %d, %d, %d, %d, %d \n', length(X), timecost, nf, ng, nH, nV, nR)
    if(isplot == 1)
        figure(1);clf
        semilogy(1 : length(F), F, '.');
        ylabel('f(x_i)');
        xlabel('iteration number');
        figure(2);clf
        semilogy(1 : length(G), G, '.');
        ylabel('|grad(x_i)|');
        xlabel('iteration number');
    end
end

function output = IsStopped(X, F, G, T, params, fns, delta)
%     fprintf('rank is %d\n', size(X{end}.U, 2));
    if(X{end}.U == X{end - 1}.U)
        output = 0;
        return;
    end
    xf = X{end};
    if(isfield(xf, 'D'))
        M = full(xf.U * sqrt(xf.D));
    else
        M = xf.U;
    end
    r = size(xf.U, 2);
    
    global isfixedrank
    if(isfixedrank)
        output = 1 - G(end);
        return;
    end
    
    idea = 1;
    if(idea == 1)
        if(r > 1)
            [U, S, V] = svd(M, 0);
            normSdsqrk = norm(diag(S)) / sqrt(size(S, 1));
            for i = 1 : size(U, 2)
                % S(i, i) / S(1, 1)
                if(S(i, i) / normSdsqrk < delta && length(F) > 10)
                    output = 1;
                    return;
                end
            end

%             if(S(end, end) / S(1, 1) < delta && length(F) > 10)
%                 output = 1;
%                 return;
%             end
            output = 0;
        else
%             if((F(end - 1) - F(end)) / F(end - 1) < 1e-6)% && length(X) > 10)
%                 output = 1;
%                 return;
%             end
            output = 1 - G(end);
            return;
        end
    elseif(idea == 2)
        if(F(end) / F(1) < 0.01)
            output = 1;
            return;
        end
        output = 0;
    elseif(idea == 3)
        if((F(end - 1) - F(end)) / F(end - 1) < 0.1)
            output = 1;
            return
        end
        output = 0;
    elseif(idea == 4)
        if(length(X) > 10)
            output = 1;
            return;
        end
        output = 0;
    end
end

function u = PhaseRetrieval_RQ(U, Sigma, b, masks)

    D = get_D_RQ(b, masks, U, Sigma);
    fns.f = @(x)f_RQ(x, b, masks, U, Sigma, D);
    fns.gf = @(x)gf_RQ(x, b, masks, U, Sigma, D);
    fns.hessianv = @(x, v)hessianv_newton_RQ(x, v, b, masks, U, Sigma, D);
    fns.R = @(x, eta)R_RQ(x, eta, U);
    fns.inpro = @inpro_RQ;
    fns.proj = @(x, eta)proj_RQ(x, eta, U);
    fns.STV = @STV;
    fns.TVaddTV = @TVaddTV;
    fns.IsStopped = @IsStopped_RQ;
    
    n = size(U, 1);
    q = randn(n, 1);
    [Q, ~] = qr([U, q], 0);
    x1.u = Q(:, end);
    
    params.x0 = x1;
    params.error = 1e-6;
    params.StopCriterion = 0;
    
    nH = 0;
    nV = 0;
    nVp = 0;
    [X, F, G, timecost, nf, ng, nH, T, nR] = RTR_Newton(fns, params);
    xend = X{end};
    u = xend.u;
end

function output = IsStopped_RQ(X, F, G)
    if(length(F)>=2 && F(end) < 0)
        output = 1;
        return;
    end
    output = 0;
    return;
end

function D = get_D_RQ(b, masks, Q, Sigma)
    [n1, n2, l] = size(masks);
    n = n1 * n2;
    m = n * l;
    k = size(Sigma, 1);
    D = zeros(m, 1);
    ZY = zeros(m, k);
    for i = 1 : l
        sZqi = 0;
        for j = 1 : k
            temp = fft2(reshape(Q(:, j), n1, n2) .* masks(:, :, i)) / sqrt(n1 * n2);
            sZqi = sZqi + Sigma(j, j) * conj(temp) .* temp;
        end
        D((i - 1) * n + 1 : i * n) = reshape(sZqi, [], 1) - b((i - 1) * n + 1 : i * n);
    end
end

function [output, x] = f_RQ(x, b, masks, U, Sigma, D)
    x.Ax = Ax_RQ(x.u, b, masks, U, Sigma, D);
    output = x.u' * x.Ax;
end

function output = Ax_RQ(v, b, masks, Q, Sigma, D)
    [n1, n2, l] = size(masks);
    n = n1 * n2;
    m = n * l;
    k = size(Sigma, 1);
    Zu = zeros(m, 1);
    for i = 1 : l
        temp = fft2(reshape(v, n1, n2) .* masks(:, :, i)) / sqrt(n1 * n2);
        Zu((i - 1) * n + 1 : i * n, 1) = reshape(temp, [], 1);
    end
    DZu = spdiags(D, 0, m, m) * Zu;
    %% Z^H *DZY
    ZHDZu = zeros(n, 1);
    for i = 1 : l
        temp = zeros(n, 1);
        temp(:, 1) = reshape(ifft2(reshape(DZu((i - 1) * n + 1 : i * n, 1) * sqrt(n1 * n2), n1, n2)), [], 1);
        temp = conj(spdiags(reshape(masks(:, :, i), [], 1), 0, n, n)) * temp;
        ZHDZu = ZHDZu + temp;
    end
    output = ZHDZu;
end

function [output, x] = gf_RQ(x, b, masks, U, Sigma, D)
    if(~isfield(x, 'Ax'))
        x.Ax = Ax_RQ(x.u, b, masks, U, Sigma, D);
    end
    temp.TV = 2 * x.Ax;
    output = proj_RQ(x, temp, U);
end

function [output, x] = hessianv_newton_RQ(x, w, b, masks, U, Sigma, D)
    % The hessian
    if(~isfield(x, 'Ax'))
        x.Ax = Ax_RQ(x.u, b, masks, U, Sigma, D);
    end
    xtax = x.u'* x.Ax;
    Aw = Ax_RQ(w.TV, b, masks, U, Sigma, D);
    temp.TV = 2 * (Aw-w.TV * xtax);
    output = proj_RQ(x, temp, U);
end

function output = inpro_RQ(x, v1, v2)
    output = real(v1.TV)' * real(v2.TV) + imag(v1.TV)' * imag(v2.TV);
end

function output = proj_RQ(x, eta, U)
    Q = [U, x.u];
    output.TV = eta.TV - Q * (Q' * eta.TV);
end

function output = R_RQ(x, eta, U)
    Q = [U, x.u + eta.TV];
    [Q, ~] = qr(Q, 0);
    output.u = Q(:, end);
end

function xnew = OneIterationSD(x0, masks, b)
    S = sqrt(x0.D);
    d = zeros(size(S));
    d(end, end) = 1;
    x1.U = x0.U * S;
    t = S(end - 1, end - 1);
    fprintf('Start one step of SD algorithm, initial stepsize:%e\n', t);
    x2.U = x0.U * (S + t * d);
    f1 = f_e(x1, masks, b);
    f2 = f_e(x2, masks, b);
    times = 0;
    while(f2 >= f1 && times < 20)
        t = t * 0.5;
        x2.U = x0.U * (S + t * d);
        f2 = f_e(x2, masks, b);
        times = times + 1;
    end
    if(times < 20)
        fprintf('Success: linesearch tried %d times. fold:%e, fnew:%e, fold - fnew:%e\n', times, f1, f2, f1 - f2);
    else
        fprinf('Warning: linesearch tried %d times, reach the max times.\n fold:%e, fnew:%e, fold - fnew:%e\n', times, f1, f2, f1 - f2);
    end
    [U, S, V] = svd(x2.U, 0);
    xnew.U = U;
    xnew.D = S * S;
    diag(xnew.D)
end

function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp] = callalgo(method, fns, params)
    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBroydenFamily(fns, params);
    elseif(method == 2)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBFGS(fns, params);
    elseif(method == 3)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRBFGS(fns, params);
    elseif(method == 4)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RSD(fns, params);
    elseif(method == 5)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RTR_SR1(fns, params);
    elseif(method == 6)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRTR_SR1(fns, params);
    elseif(method == 7)
        [X, F, G, timecost, nf, ng, T, nR] = RTR_SD(fns, params);
    elseif(method == 8)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RGS(fns, params);
    elseif(method == 9)
        [X, F, G, timecost, nf, ng, nH, T, nR] = RTR_Newton(fns, params);
    elseif(method == 10)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RWRBFGS(fns, params);
    elseif(method == 11)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RCG(fns, params);
    elseif(method == 12)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RNewton(fns, params);
    elseif(method == 13)
        [X, F, G, timecost, nf, ng, T, nR] = check_grad_Hess(fns, params);
    end
end

function output = STV(x, a, eta)
    output.TV = a * eta.TV;
end
    
function output = TVaddTV(x, v1, v2)
    output.TV = v1.TV + v2.TV;
end

function Bc = make_quotient_basis(x)
    n = size(x.U, 1);
    nbc = x.U * complex(0, 1);
    nbr = [real(nbc);imag(nbc)];
    Br = null(nbr');
    Bc = complex(Br(1 : n, :), Br(n + 1 : 2 * n, :));
end

function [output, x] = f_e(x, masks, b, kappa)
    [n1, n2, l] = size(masks);
    n = n1 * n2;
    k = size(x.U, 2);
    m = l * n;
    D = zeros(m, 1);
    ZY = zeros(m, k);
    for i = 1 : l
        sZqi = zeros(n, 1);
        for j = 1 : k
            temp = reshape(fft2(reshape(x.U(:, j), n1, n2) / sqrt(n1 * n2) .* masks(:, :, i)), [], 1);
            ZY((i - 1) * n + 1 : i * n, j) = reshape(temp, [], 1);
            sZqi = sZqi + conj(temp) .* temp;
        end
        D((i - 1) * n + 1 : i * n) = reshape(sZqi, [], 1) - b((i - 1) * n + 1 : i * n);
    end
    output = norm(D)^2 / norm(b, 'fro')^2;
    x.ZY = ZY;
    x.cD = D;
    
    output = output + kappa * norm(x.U, 'fro')^2;
end

%% quotient space using Euclidean metric=====================================================
function [output, x] = f_q(x, masks, b, kappa)
    [n1, n2, l] = size(masks);
    n = n1 * n2;
    k = size(x.U, 2);
    m = l * n;
    D = zeros(m, 1);
    ZY = zeros(m, k);
    for i = 1 : l
        sZqi = zeros(n, 1);
        for j = 1 : k
            temp = reshape(fft2(reshape(x.U(:, j), n1, n2) / sqrt(n1 * n2) .* masks(:, :, i)), [], 1);
            ZY((i - 1) * n + 1 : i * n, j) = reshape(temp, [], 1);
            sZqi = sZqi + conj(temp) .* temp;
        end
        D((i - 1) * n + 1 : i * n) = reshape(sZqi, [], 1) - b((i - 1) * n + 1 : i * n);
    end
    output = norm(D)^2 / norm(b, 'fro')^2;
    x.ZY = ZY;
    x.cD = D;
    if(k ~= 1)
        output = output + kappa * norm(x.U, 'fro')^2;%% add penelty term
    end
%     clearvars -except output x
end

function [output, x] = gf_q(x, masks, b, kappa)
    [n1, n2, l] = size(masks);
    n = n1 * n2;
    m = l * n;
    k = size(x.U, 2);
%     x.ZY;
%     x.cD;
    if(~isfield(x, 'ZY') || ~isfield(x, 'cD'))
        D = zeros(m, 1);
        ZY = zeros(m, k);
        for i = 1 : l
            sZqi = 0;
            for j = 1 : k
                temp = fft2(reshape(x.U(:, j), n1, n2) .* masks(:, :, i)) / sqrt(n1 * n2);
                ZY((i - 1) * n + 1 : i * n, j) = reshape(temp, [], 1);
                sZqi = sZqi + conj(temp) .* temp;
            end
            D((i - 1) * n + 1 : i * n) = reshape(sZqi, [], 1) - b((i - 1) * n + 1 : i * n);
        end
        x.cD = D;
        x.ZY = ZY;
    end
    D = x.cD;
    ZY = x.ZY;
    DZY = zeros(m, k);
    for i = 1 : k
        DZY(:, i) = D.*ZY(:, i);
    end
    %% Z^H *DZY
    Y = zeros(n, k);
    for i = 1 : l
        temp = zeros(n, k);
        for j = 1 : k
            temp(:, j) = reshape(ifft2(reshape(DZY((i - 1) * n + 1 : i * n, j) * sqrt(n1 * n2), n1, n2)), [], 1);
            temp(:, j) = conj(reshape(masks(:, :, i), [], 1)) .* temp(:, j);
        end
        Y = Y + temp;
    end
    YY.TV = 4 * Y / norm(b, 'fro')^2;
    if(k ~= 1)
        YY.TV = YY.TV + kappa * 2 * x.U;%% for penelty term
    end
    
    x.UtU = x.U' * x.U;
    
    output.TV = YY.TV / x.UtU;

%     exh1 = output.TV;
%     exnorm = trace((x.U' * x.U) * (output.TV' * output.TV))
    x.gf = output.TV;
    output = GetIntr(x, output);
    
%     innorm = output.TV(:)'*output.TV(:)
    
%     %%----
%     inh1 = output.TV;%%---
%     
%     exoutput = GetExtr(x, output);
%     exh2 = exoutput.TV;%%---
%     inoutput = GetIntr(x, exoutput);
%     inh2 = inoutput.TV;%%---
%     
%     norm(exh1 - exh2)
%     norm(inh1 - inh2)
    
    
%     error('asf');%%----
%     clearvars -except output x
end

function output = symm(M)
    output = (M + M') / 2;
end

function output = R_Exp_q(x, eta)
    eta = GetExtr(x, eta);
    output.U = x.U + eta.TV;
    output = QRmatrices(output);
    [Ubasis, output] = xUbasis(output);
    output.Ubasis = Ubasis;
end

function x = QRmatrices(x)
    [n, k] = size(x.U);
    HHR = zeros(n, k);
    taus = zeros(k, 1);
    A = x.U;
    
    for i = 1 : k
        HHR(i : n, i) = A(i : n, i);
        HHR(i, i) = HHR(i, i) + sign(A(i, i)) * norm(A(i : n, i));
        taus(i) = 2 / (HHR(i : n, i)' * HHR(i : n, i));
        A = A - taus(i) * HHR(:, i) * (HHR(:, i)' * A);
    end
    x.HHR = HHR;
    x.taus = taus;
end

function output = HHRTeta(x, eta)
    HHR = x.HHR;
    taus = x.taus;
    [n, k] = size(x.U);
    output = eta.TV;
    for i = 1 : k
        output = output - taus(i) * HHR(:, i) * (HHR(:, i)' * output);
    end
end

function output = HHReta(x, eta)
    HHR = x.HHR;
    taus = x.taus;
    [n, k] = size(x.U);
    output = eta.TV;
    for i = k : -1 : 1
        output = output - taus(i) * HHR(:, i) * (HHR(:, i)' * output);
    end
end

function output = GetIntr(x, eta)
    [n, k] = size(x.U);
    if(~isfield(x, 'HHR'))
        x = QRmatrices(x);
    end
    MnhSK = HHRTeta(x, eta);
    if(~isfield(x, 'L'))
        x.L = chol(x.U' * x.U, 'lower');
    end
    L = x.L;
    UHV = (L\(x.U' * eta.TV));
    UHV = UHV * L;
    output.TV = zeros(k * k + k * (n - k), 1);
    idx = 1;
    factor = 1;
    for i = 1 : k
        output.TV(idx) = real(UHV(i, i)) / factor;
        idx = idx + 1;
    end
    r2 = sqrt(2);
    for i = 1 : k
        for j = i + 1 : k
            output.TV(idx) = real(UHV(i, j)) * r2 / factor;
            idx = idx + 1;
            output.TV(idx) = imag(UHV(i, j)) * r2 / factor;
            idx = idx + 1;
        end
    end
    MnhSK = MnhSK * L;
    output.TV(k*k+1:k*k + k * (n - k), 1) = reshape(MnhSK(k+1:n, :), [], 1);
end

function output = GetExtr(x, ineta)
    [n, k] = size(x.U);
    if(~isfield(x, 'HHR'))
        x = QRmatrices(x);
    end
    if(~isfield(x, 'L'))
        x.L = chol(x.U' * x.U, 'lower');
    end
    L = x.L;
    S = zeros(k, k);
    r2 = sqrt(2);
    factor = 1;%sqrt(x.U(:)'*x.U(:));
    idx = 1;
    for i = 1 : k
        S(i, i) = ineta.TV(i) * factor;
        idx = idx + 1;
    end
    for i = 1 : k
        for j = i + 1 : k
            S(i, j) = complex(ineta.TV(idx), ineta.TV(idx + 1)) / r2 * factor;
            S(j, i) = S(i, j)';
            idx = idx + 2;
        end
    end
    output.TV = x.U * (L' \ S);
    eta.TV = zeros(n, k);
    eta.TV(k + 1 : n, :) = reshape(ineta.TV(k*k+1:end, 1), n - k, k);
    YpK = HHReta(x, eta);
    output.TV = output.TV + YpK;
    output.TV = output.TV / L;
end

function [output, x] = xUbasis(x)
    [n, k] = size(x.U);
    M = x.U'*x.U;
    Mnh = M^(-0.5);
    x.Mnh = Mnh;
    RMnh = real(Mnh);
    IMnh = imag(Mnh);
    B = zeros(2 * k * k, k * k);
    col = 1;
    for i = 1 : k
        for j = i : k
            B((j - 1) * k + 1 : j * k, col) = RMnh(:, i);
            B((j - 1) * k + 1 + k * k : j * k + k * k, col) = IMnh(:, i);
            B((i - 1) * k + 1 : i * k, col) = RMnh(:, j);
            B((i - 1) * k + 1 + k * k: i * k + k * k, col) = IMnh(:, j);
            col = col + 1;
        end
    end
    for i = 1 : k
        for j = i + 1 : k
            B((j - 1) * k + 1 : j * k, col) = - IMnh(:, i);
            B((j - 1) * k + 1 + k * k : j * k + k * k, col) = RMnh(:, i);
            B((i - 1) * k + 1 : i * k, col) = IMnh(:, j);
            B((i - 1) * k + 1 + k * k: i * k + k * k, col) = -RMnh(:, j);
            col = col + 1;
        end
    end
    [output, ~] = qr(B, 0);
end

function output = inpro_q(x, v1, v2)
    output = sum(sum(real(v1.TV) .* real(v2.TV))) + sum(sum(imag(v1.TV) .* imag(v2.TV)));
end

function output = Hv_q(x, H, v)
    [n, k] = size(v.TV);
    output.TV = reshape(H * reshape(v.TV, [], 1), n, k);
end

function output = phi_q()
    output = 0;
end

function output = beta_q(x1, eta, x2)
    output = 1;
end


function [output, d, x, y] = TranvProj_q(x, d, y, v)
    eta = GetExtr(x, v);
    yy = y.U' * y.U;
    ye = y.U' * eta.TV;
    ye = ye - ye';
    omega = lyap(yy, yy, -ye);
    output.TV = eta.TV - y.U * omega;
    output = GetIntr(y, output);
    return;
end

function [output, d, x, y] = Tranv_q(x, d, y, v)
    output = v;
end

function [output, R] = qf_thin(B)
    k = size(B, 2);
    [output, R] = qr(B, 0);
    diagR = diag(R);
    output = output(:, abs(diagR) > 1e-10);
    output = output * spdiags(sign(diagR), 0, size(output, 2), size(output, 2));
end

function output = build_perp_basis(x)
    [n, p] = size(x.U);
    col = 1;
    RU = real(x.U);
    IU = imag(x.U);
    Basis = zeros(2 * n * p, p * p);
    for i = 1 : p
        Basis((i - 1) * n + 1 : i * n, col) = - IU(:, i);
        Basis((i - 1) * n + 1 + p * n : i * n + p * n, col) = RU(:, i);
        col = col + 1;
    end
    for i = 2 : p
        for j = i : p
            Basis((i - 2) * n + 1 : (i - 1) * n, col) = - RU(:, j);
            Basis((i - 2) * n + 1 + p * n : (i - 1) * n + p * n, col) = - IU(:, j);
            Basis((j - 1) * n + 1 : j * n, col) = RU(:, i - 1);
            Basis((j - 1) * n + 1 + p * n : j * n + p * n, col) = IU(:, i - 1);
            col = col + 1;
            Basis((i - 2) * n + 1 : (i - 1) * n, col) = - IU(:, j);
            Basis((i - 2) * n + 1 + p * n : (i - 1) * n + p * n, col) = RU(:, j);
            Basis((j - 1) * n + 1 : j * n, col) = - IU(:, i - 1);
            Basis((j - 1) * n + 1 + p * n : j * n + p * n, col) = RU(:, i - 1);
            col = col + 1;
        end
    end
    output.B = Basis;
    col = 1;
    M = x.U' * x.U;
    M = M^(0.5);
    RM = real(M);
    IM = imag(M);
    MBasis = zeros(2 * p * p, p * p);
    for i = 1 : p
        MBasis((i - 1) * p + 1 : i * p, col) = - IM(:, i);
        MBasis((i - 1) * p + 1 + p * p : i * p + p * p, col) = RM(:, i);
        col = col + 1;
    end
    for i = 2 : p
        for j = i : p
            MBasis((i - 2) * p + 1 : (i - 1) * p, col) = - RM(:, j);
            MBasis((i - 2) * p + 1 + p * p : (i - 1) * p + p * p, col) = - IM(:, j);
            MBasis((j - 1) * p + 1 : j * p, col) = RM(:, i - 1);
            MBasis((j - 1) * p + 1 + p * p : j * p + p * p, col) = IM(:, i - 1);
            col = col + 1;
            MBasis((i - 2) * p + 1 : (i - 1) * p, col) = - IM(:, j);
            MBasis((i - 2) * p + 1 + p * p : (i - 1) * p + p * p, col) = RM(:, j);
            MBasis((j - 1) * p + 1 : j * p, col) = - IM(:, i - 1);
            MBasis((j - 1) * p + 1 + p * p : j * p + p * p, col) = RM(:, i - 1);
            col = col + 1;
        end
    end
    BTB = MBasis'*MBasis;
    output.BTB = BTB;
    output.M = inv(BTB);
return;
end

function [output, d, x, y] = invTranv_q(x, d, y, v)
    output = v;
end

function output = TranH_q(x, d, y, H)
    output = H;
end

function output = rank1operator_q(x, v1, v2)
    output = reshape(v1.TV, [], 1) * reshape(v2.TV, [], 1)';
end

function output = operadd_q(x, H1, H2)
    output = H1 + H2;
end

%% right invariant metric=============================================
function [output, x] = f(x, masks, b)
    [n1, n2, l] = size(masks);
    n = n1 * n2;
    m = n * l;
    Q = x.U;
    Sigma = x.D;
    k = size(Sigma, 1);
    D = zeros(m, 1);
    ZY = zeros(m, k);
    for i = 1 : l
        sZqi = 0;
        for j = 1 : k
            temp = fft2(reshape(Q(:, j), n1, n2) .* masks(:, :, i)) / sqrt(n1 * n2);
            ZY((i - 1) * n + 1 : i * n, j) = reshape(temp * sqrt(Sigma(j, j)), [], 1);
            sZqi = sZqi + Sigma(j, j) * conj(temp) .* temp;
        end
        D((i - 1) * n + 1 : i * n) = reshape(sZqi, [], 1) - b((i - 1) * n + 1 : i * n);
    end
    x.ZY = ZY;
    x.cD = D;
    output = norm(D)^2 / norm(b, 'fro')^2;
end

function [output, x] = gf(x, masks, b)
    ZY = x.ZY;
    D = x.cD;
    [n1, n2, l] = size(masks);
    n = n1 * n2;
    m = l * n;
    Q = x.U;
    Sigma = x.D;
    k = size(Sigma, 1);
    DZY = spdiags(D, 0, m, m) * ZY;
    %% Z^H *DZY
    zeta = zeros(n, k);
    for i = 1 : l
        temp = zeros(n, k);
        for j = 1 : k
            temp(:, j) = reshape(ifft2(reshape(DZY((i - 1) * n + 1 : i * n, j) * sqrt(n1 * n2), n1, n2)), [], 1);
        end
        temp = conj(spdiags(reshape(masks(:, :, i), [], 1), 0, n, n)) * temp;
        zeta = zeta + temp;
    end
    Voutput.TV = 4 * zeta * Sigma / norm(b, 'fro')^2;
    output = proj(x, Voutput);
    return;
end

function output = fhessianv(x, v, masks, b)
    Q = x.U;
    Sigma = x.D;
    [n1, n2, l] = size(masks);
    k = size(Sigma, 1);
    n = n1 * n2;
    m = n * l;
    YYI = spdiags(1./diag(Sigma), 0, k, k);
    YY = Sigma;
    Y = Q * sqrt(Sigma);
    vYYv = v.TV' * Y;
    vYYv = vYYv + vYYv';
    D = zeros(m, 1);
    ZY = zeros(m, k);
    Zv = zeros(m, k);
    for i = 1 : l
        sZqi = 0;
        for j = 1 : k
            temp = fft2(reshape(Q(:, j), n1, n2) .* masks(:, :, i)) / sqrt(n1 * n2);
            ZY((i - 1) * n + 1 : i * n, j) = reshape(temp * sqrt(Sigma(j, j)), [], 1);
            Zv((i - 1) * n + 1 : i * n, j) = reshape(fft2(reshape(v.TV(:, j), n1, n2) .* masks(:, :, i)) / sqrt(n1 * n2), [], 1);
            sZqi = sZqi + Sigma(j, j) * conj(temp) .* temp;
        end
        D((i - 1) * n + 1 : i * n) = reshape(sZqi, [], 1) - b((i - 1) * n + 1 : i * n);
    end
    DD = ZY .* conj(Zv);
    if(k ~= 1)
        DD = sum(DD')';
    end
    DD = 2 * real(DD);
    DZYvYYv = full(full(spdiags(D, 0, m, m) * ZY) * vYYv);
    DZvYY = full(full(spdiags(D, 0, m, m) * Zv) * Sigma);
    DDZYYY = full(full(spdiags(DD, 0, m, m) * ZY) * Sigma);
    DZALL = DZYvYYv + DZvYY + DDZYYY;
    DZYYY = full(full(spdiags(D, 0, m, m) * ZY) * Sigma);
    %% Z^H *DZALL and Z^H DZYYY
    zeta = zeros(n, k);
    dotzeta = zeros(n, k);
    for i = 1 : l
        temp = zeros(n, k);
        temp1 = zeros(n, k);
        for j = 1 : k
            temp(:, j) = reshape(ifft2(reshape(DZALL((i - 1) * n + 1 : i * n, j) * sqrt(n1 * n2), n1, n2)), [], 1);
            temp1(:, j) = reshape(ifft2(reshape(DZYYY((i - 1) * n + 1 : i * n, j) * sqrt(n1 * n2), n1, n2)), [], 1);
        end
        temp = spdiags(reshape(masks(:, :, i), [], 1), 0, n, n)' * temp;
        dotzeta = dotzeta + temp;
        temp1 = spdiags(reshape(masks(:, :, i), [], 1), 0, n, n)' * temp1;
        zeta = zeta + temp1;
    end
    dotzeta = 4 * dotzeta;
    zeta = 4 * zeta;
    
    temp1 = spdiags(1./diag(Sigma), 0, k, k) * (Y' * zeta);
    temp1 = temp1 + temp1';
    v1 = diag(Sigma);
    v2 = 1./diag(Sigma);
    temp2 = v2 * v1' + v1 * v2';
    M = temp1 ./ temp2;
    
    M1 = kron(Sigma, YYI) + kron(YYI, Sigma);
    invM1 = spdiags(1./diag(M1), 0, k*k, k*k);
    temp1 = YYI * (Y' * zeta);
    M2 = temp1 + temp1';
    
    temp2 = - YYI * vYYv * YYI * (Y' * zeta) + YYI * (v.TV' * zeta + Y' * dotzeta);
    temp2 = temp2 + temp2';
    temp3 = kron(vYYv, YYI) + kron(Sigma, - YYI * vYYv * YYI);
    temp4 = kron(YYI, vYYv) + kron(-YYI * vYYv * YYI, Sigma);
    dotM = reshape(invM1 * reshape(temp2, [], 1) - (invM1 * ((temp3 + temp4) * (invM1 * reshape(M2, [], 1)))), k, k);
    
    Dgradf = v.TV * (YYI * M * Sigma) - Y * (YYI * vYYv * YYI * M * Sigma) ...
        + Y * (YYI * M * vYYv) + Y * (YYI * dotM * Sigma) ...
        + dotzeta - Q * (Q' * dotzeta) - v.TV * (YYI * (Y' * zeta)) - Y * (YYI * (v.TV' * zeta)) ...
        + Y * (YYI * vYYv * YYI * (Y' * zeta));
    Vzeta.TV = zeta;
    gradf = proj(x, Vzeta);    
    temp5 = Y * (YYI * (gradf.TV' * v.TV)) - v.TV * (YYI * (gradf.TV' * Y));
    temp6 = Y * (YYI * (v.TV' * gradf.TV)) - gradf.TV * (YYI * (v.TV' * Y));
    temp7 = gradf.TV * (YYI * (Y' * v.TV)) + v.TV * (YYI * (Y' * gradf.TV));
    Voutput.TV = (Dgradf + 0.5 * (temp5 + temp6 - temp7)) / norm(b, 'fro')^2;
    output = proj(x, Voutput);
    return;

end

function output = proj(x, eta)
    Q = x.U;
    Sigma = x.D;
    k = size(Sigma, 1);
    temp1 = spdiags(1./diag(Sigma), 0, k, k) * (Q * sqrt(Sigma))' * eta.TV;
    temp1 = temp1 + temp1';
    v1 = diag(Sigma);
    v2 = 1./diag(Sigma);
    temp2 = v2 * v1' + v1 * v2';
    M = temp1 ./ temp2;
    output.TV = Q * (sqrt(Sigma) * (M .* (v2 * v1'))) + eta.TV - Q * (Q' * eta.TV);
end

function output = R_Exp(x, eta)
    [M, K] = getMK(x, eta);
    Q = x.U;
    Sigma = x.D;
    k = size(Sigma, 1);
    YYH = sqrt(Sigma);
    YYHI = spdiags(1./diag(YYH), 0, k, k);
    Omega = YYHI * M * YYH - YYH * M * YYHI;
    S = (YYHI * K') * (K * YYHI);
    X = expm(full([Omega, -S; eye(k), zeros(k, k)]));
    Temp = (Q * X(1:k, 1:k) + K * (YYHI * X(k+1:2*k,1:k))) * YYH * expm(M);
    [U, D, V] = svd(full(Temp), 0);
    output.U = U;
    output.D = sparse(D).^2;
end

function output = inpro(x, v1, v2)
    Q = x.U;
    Sigma = x.D;
    k = size(Sigma, 1);
    YYI = spdiags(1./diag(Sigma), 0, k, k);
    temp1 = full(v1.TV * YYI);
    temp2 = v2.TV;
    output = sum(sum(real(temp1) .* real(temp2))) + sum(sum(imag(temp1) .* imag(temp2)));
end


%%--------------------------------------------------------------------------

function Z = makeZ(masks)
    [n1, n2, l] = size(masks);
    n = n1 * n2;
    m = n * l;
    Z = zeros(m, n);
    F1 = fourierM(n1);
    F2 = fourierM(n2);
    
    KF = kron(F2, F1);
    for i = 1 : l
        Z((i - 1) * n + 1 : i * n, :) = KF * spdiags(reshape(masks(:, :, i), [], 1), 0, n, n);
    end
end

function F = fourierM(n)
    F = zeros(n, n);
    w = complex(cos(- 2 * pi / n), sin(- 2 * pi / n));
    for i = 1 : n
        for j = 1 : n
            F(i, j) = w^((i - 1) * (j - 1));
        end
    end
end

function [M, K] = getMK(x, eta)
    Q = x.U;
    Sigma = x.D;
    k = size(Sigma, 1);
    temp1 = spdiags(1./diag(Sigma), 0, k, k) * (Q * sqrt(Sigma))' * eta.TV;
    temp1 = temp1 + temp1';
    v1 = diag(Sigma);
    v2 = 1./diag(Sigma);
    temp2 = v2 * v1' + v1 * v2';
    M = temp1 ./ temp2;
    K = eta.TV - Q * (Q' * eta.TV);
end
