function varargout = RTR_Newton(fns,params)
% Solver of Riemannian trust region with Newton method
%
% INPUT:
% fns : a struct that contains required function handles
% required:
%     fns.f(x) : return objective function value at x.
%     fns.gf(x) : return the gradient of objection function at x.
%     fns.fhess(x) : return the gradient of objection function at x.
% not required if params.manifold, params.retraction and params.vector_transport is provided.
%     fns.STV(a, v) : return a tangent vector which is a times v.
%     fns.TVaddTV(v1, v2) : return a tangent vector which is equal to v1 adds v2.
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.proj(x, eta) : return P_x(eta) by projecting v to tangent space of x.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
% 
% params : a struct that contains parameters that are used in the solver.
% required:
%     params.x0 : initial approximation of minimizer.
% not required:
%     params.error [1e-5]       : tolerance of stopping criterion.
%     params.StopCriterion [3]  : stopping criterion, 0 means call fns.IsStopped to check
%                                 1 means stop when relative error of objective function is less than tolerance,
%                                 2 means stop when norm of gradient is less than tolerance,
%                                 3 means stop when norm of gradient over initial norm of gradient is less than tolerance,
%     params.max_t [300]        : the maximum number of iterations.
%     params.debug [1]          : '0' means almost silence, '1' means some output. '2' means a lot of output, '3' means more than need to know.
%     params.c [0.1]            : acceptance or rejection constant
%     params.tau1 [0.25]        : shrink radius constant
%     params.tau2 [2]           : expansion radius constant
%     params.max_Delta [200]    : max of Delta
%     params.min_Delta [1e-10]   : min of Delta
%     params.theta[1]         : A stopping criterion parameter for solving local model
%     params.kappa[0.1]         : A stopping criterion parameter for solving local model
%     params.min_innit[0]       : the minimum number of iterations for solving local model
%     params.max_innit[200]     : the maxmum number of iterations for solving local model
% The following give some examples of manifold. For each manifold, some retractions and vector transports are provided.
%     params.manifold : the manifold which objective function is on.
%         '1', sphere : S^{n - 1}
%         '2', Stiefel manifold : St(p, n)
%         '3', orthogonal group : O(n)
%         '4', grassmann manifold : Gr(p, n)
%
% OUTPUT:
% varargout : contains output parameters
%     varargout{1} : all iterates
%     varargout{2} : function values of all iterates
%     varargout{3} : norm of gradient of all iterates
%     varargout{4} : total computational time
%     varargout{5} : number of function evaluations
%     varargout{6} : number of gradient evaluations
%     varargout{7} : number of Hessian actions
%     varargout{8} : accumulated computational time for iterations
%     varargout{9} : number of retraction actions
% 
% By Wen Huang, the solver of local model is a modified version of that in http://www.math.fsu.edu/~cbaker/GenRTR/

    fprintf('RTR_Newton\n')
    if nargin < 2,
       error('Invalid arguments: the number of arguments is invalid.');
    end
    % set default parameters and check the legal range of parameters
    params = set_default_para(params, 'StopCriterion', 3, 'int', 0, 3);
    params = set_default_para(params, 'error', 1e-6, 'float', 0, 1);
    params = set_default_para(params, 'max_t', 300, 'int', 1, inf);
    params = set_default_para(params, 'debug', 1, 'int', 0, 3);
    params = set_default_para(params, 'c', 0.1, 'float', 0, 0.25);
    params = set_default_para(params, 'tau1', 0.25, 'float', 0.1, 0.9);
    params = set_default_para(params, 'tau2', 2, 'float', 1.5, 5);
    params = set_default_para(params, 'max_Delta', 200, 'float', 1, inf);
    params = set_default_para(params, 'min_Delta', 1e-10, 'float', 0, 1);
    params = set_default_para(params, 'theta', 1, 'float', 0, 1);
    params = set_default_para(params, 'kappa', 0.1, 'float', 0, 1);
    params = set_default_para(params, 'min_innit', 0, 'int', 0, inf);
    params = set_default_para(params, 'max_innit', 200, 'int', 0, inf);
    params = set_default_para(params, 'keepnumX', 2100000000, 'int', 1, inf);
    params = set_default_para(params, 'useRand', 0, 'int', 0, 1);
    params = set_default_para(params, 'isplot', 0, 'int', 0, 1);
    if(isfield(params, 'manifold') && isfield(params, 'retraction') && isfield(params, 'vector_transport'))
        fs = choose_manifold(params.manifold, params.retraction, params.vector_transport);
        fns.R = fs.R;
        fns.proj = fs.proj;
        fns.inpro = fs.inpro;
        fns.STV = fs.STV;
        fns.TVaddTV = fs.TVaddTV;
        fns.STV = fs.STV;
        fns.TVaddTV = fs.TVaddTV;
    end
    % Check whether function handles are legal
    check_function_handle(fns, 'f');
    check_function_handle(fns, 'gf');
    check_function_handle(fns, 'hessianv');
    check_function_handle(fns, 'R');
    check_function_handle(fns, 'proj');
    check_function_handle(fns, 'inpro');
    check_function_handle(fns, 'STV');
    check_function_handle(fns, 'TVaddTV');
    if(params.StopCriterion == 0)
        check_function_handle(fns, 'IsStopped');
    end
    % Check whether required parameters are legal
    if(~isfield(params, 'x0'))
        error('Invalid arguments: missing initial x0');
    end

    fns.haveprec = 0;
    fns.g = fns.inpro;
    % Initialization
    c = params.c;
    tau1 = params.tau1;
    tau2 = params.tau2;
    err = inf;
    err_v = inf;
    times = 0;
    Delta = 1;
    x = params.x0;
    err_g = 1;
    err_r = 1;
    [fx, x] = fns.f(x);
    [gradf, x] = fns.gf(x);
%     B = params.B0;
    
    X{1} = x;
    F(1) = fx;
    G(1) = sqrt(fns.inpro(x, gradf, gradf));

    nf = 1;
    ng = 1;
    nH = 0;
    nV = 0;
    nR = 0;
    ind = 1;
    Grads = [];
    updateinfo = '';
    timeinfo = '';
    fprintf('times : %d, f(x) : %e\n', times, fx);
    tic
    T(1) = 0;
    while (err > params.error && times < params.max_t)
        if(params.debug > 1)
            tCGtimestart = toc;
        end
        EMPTYTV = fns.STV(x, 0, gradf);
        [eta, inner_it, stop_tCG, timeinfo] = tCG_TR(fns, x, gradf, EMPTYTV, Delta, times, params, timeinfo, 'Newton');
        if(params.debug > 1)
            timeinfo = sprintf('time, tCG:%.2e,', toc - tCGtimestart);
            getftimestart = toc;
        end
        xtemp = fns.R(x, eta);
        [fxtemp, xtemp] = fns.f(xtemp);
        if(params.debug > 1)
            timeinfo = [timeinfo sprintf('get f:%.2e,', toc - getftimestart)];
            checkacceptencetimestart = toc;
        end
        
        Hvtemp = fns.hessianv(x, eta); %Hv(x, B, eta);
%         rho = (fx - fxtemp) / (- fns.inpro(x, gradf, eta) - 0.5 * fns.inpro(x, eta, Hvtemp));
        rho = (fx - fxtemp) / max(eps, (- fns.inpro(x, fns.TVaddTV(x, gradf,  fns.STV(x, 0.5, Hvtemp)), eta)));
        if(rho > 0.75)
            if(stop_tCG == 2 || stop_tCG == 1)
                Delta = tau2 * Delta;
            end
            if(Delta > params.max_Delta)
                fprintf('reach max of Delta\n');
                break;
            end
        elseif(rho < 0.25)
            Delta = tau1 * Delta;
            if(Delta < params.min_Delta)
                fprintf('reach min Delta\n');
                break;
            end
        end
        if(rho > c || (abs((fx - fxtemp)/(abs(fx) + 1)) < sqrt(eps) && fxtemp < fx))
            err_r = abs((fx - fxtemp) / fxtemp);
            x = xtemp;
            [gradf, x] = fns.gf(x);%gradftemp;
            fx = fxtemp;
            err_g = sqrt(fns.inpro(x, gradf, gradf));
            if(params.debug > 0)
                updateinfo = '';
            end
        else
            if(params.debug > 0)
                updateinfo = ('without updating x_k');
            end
        end
        if(params.debug > 1)
            timeinfo = [timeinfo sprintf('check acceptence f:%.2e,', toc - checkacceptencetimestart)];
        end
        times = times + 1;
        nf = nf + 1;
        ng = ng + 1;
        nR = nR + 1;
        nH = nH + 1 + inner_it;
        X{times + 1} = x;
        if(times + 1 - params.keepnumX > 0)
            X{times + 1 - params.keepnumX} = [];
        end
        F(times + 1) = fx;
        G(times + 1) = err_g;
        T(times + 1) = toc;
        if(params.StopCriterion == 0)
            err = 1 - fns.IsStopped(X, F, G, T, params, fns);
        elseif(params.StopCriterion == 1)
            err = err_r;
        elseif(params.StopCriterion == 2)
            err = err_g;
        elseif(params.StopCriterion == 3)
            err = err_g / G(1);
        end
        if(params.debug > 1)
            cprintf('k', ['\t' timeinfo '\n']);
        end
        if(params.debug > 0)
            if(isempty(updateinfo))
                fprintf('i:%d,f:%.3e,df/f:%.3e,|gf|:%.3e,radius:%.3e,time:%.2e,stoptCG:%d,in_it:%d,rho:%.2e\n', ...
                    times, fx, err_r, err_g, Delta, T(end), stop_tCG, inner_it, rho);
            else
                fprintf(['i:%d, ' updateinfo ', radius reduces to %.3e\n'], times, Delta);
            end
        end
    end
    timecost = toc;
    fprintf('Num. of Iter.: %d, time cost: %f seconds, Num. of Fun, Gra, Hv, VT and R: %d, %d, %d, %d, %d \n', times, timecost, nf, ng, nH, nV, nR)
    fprintf('f: %e, |gradf|: %e, |gradf/gradf0|: %e \n', F(end), G(end), G(end)/G(1))
    if(params.isplot)
        figure(1);clf
        scatter(1 : length(F), F, '.');
        ylabel('f(x_i)');
        xlabel('iteration number');
        figure(2);clf
        semilogy(1 : length(G), G, '.');
        ylabel('|grad(x_i)|');
        xlabel('iteration number');
    end
    varargout{1} = X;
    varargout{2} = F;
    varargout{3} = G;
    varargout{4} = timecost;
    varargout{5} = nf;
    varargout{6} = ng;
    varargout{7} = nH;
    varargout{8} = T;
    varargout{9} = nR;
end